const datauri = require('datauri/sync');
const isLocalPath = require('is-local-path');
const isTemplateExpression = require('./is-template-expression');
const path = require('path');
const postcss = require('postcss');
const postcssUrl = require('postcss-url');
const R = require('ramda');
const url = require('url');

const collapseWhitespace = (str) => {
		let s = str.replace(/[\s\xa0]+/g, ' ').replace(/^\s+|\s+$/g, '');
		return s;
}
const forEachIndexed = R.addIndex(R.forEach);
const resolve = path.resolve;

const augmentError = (error, filename, files) => {
	if (!error.filename) error.filename = filename;
	error.files = R.uniq(R.concat(files, error.files || []));
	return error;
};

/**
 * Returns url path without query string and hash if present.
 *
 * @param path
 * @returns path
 */
const cleanUrl = R.pipe(
	url.parse,
	R.pick(['protocol', 'host', 'pathname']),
	url.format,
	decodeURI
);
/**
 * Convert local url data type paths to datauris.
 *
 * @param css
 * @param filename
 * @returns {{css: (css|any), files: Array}}
 */
const inlineUrl = async (filename, css) => {
	const basePath = path.dirname(filename);
	var files = [];
	const inline = url => {
		try {
			if (isLocalPath(url.url) && !isTemplateExpression(url.url)) {
				url.url = cleanUrl(url.url);
				url = resolve(basePath, url.url);
				files = R.append(url, files);
				try{
					url = datauri(url);
				}
				catch(e){
					throw e;
				}
				return url.content;
			}
			else{
				return url.url;
			}
			
		}
		catch (error) { throw augmentError(error, filename, files); }
	};
	try{
		css = await postcss()
		.use(postcssUrl({ url: inline }))
		.process(css);
	files = R.uniq(files);
	return { css: css.css, files };
	}
	catch(e){
		throw e;
	}	
};

const inlineStyles = async ($, filename) => {
	var files = [];
	try {
		const styles = $('style')
			.toArray();
		let contents = [];
		await Promise.all(styles.map(async (style)=>{
			const css = $(style).html();
			const result = await inlineUrl(filename, css);
			files = R.concat(files, result.files);
			contents.push(result.css);
		}));
		// const contents = R.map(style => {
		// 	const css = $(style).html();
		// 	const result = await inlineUrl(filename, css);
		// 	files = R.concat(files, result.files);
		// 	return result.css;
		// }, styles);
		const replaceStyle = (style, index) => $(style).html(contents[index]);
		forEachIndexed(replaceStyle, styles);
		return { $, files };
	}
	catch (error) { throw augmentError(error, filename, files); }
};

const prefix = 'selector {';
const suffix = '}';
const matchStyle = new RegExp(`^${prefix}\\s*(.*)\\s*${suffix}$`);
const wrap = style => `${prefix}${style}${suffix}`;
const unwrap = rule => rule.replace(matchStyle, '$1');

const inlineStyleAttributes = async ($, filename) => {
	var files = [];
	try {
		const elements = $('*')
			.filter('[style]')
			.toArray();

		let styles = [];
		await Promise.all(elements.map(async (element)=>{
				var style = $(element).attr('style');
				const rule = wrap(style);
				const result = await inlineUrl(filename, rule);
				files = R.concat(files, result.files);
				style = R.pipe( collapseWhitespace, unwrap )(result.css);
				styles.push(style);
			}));
		// const styles = R.map(element => {
		// 	var style = $(element).attr('style');
		// 	const rule = wrap(style);
		// 	const result = inlineUrl(filename, rule);
		// 	files = R.concat(files, result.files);
		// 	style = R.pipe( collapseWhitespace, unwrap )(result.css);
		// 	return style;
		// }, elements);
		const replaceElementStyle = (element, index) => $(element).attr('style', styles[index]);
		forEachIndexed(replaceElementStyle, elements);
		return { $, files };
	}
	catch (error) { throw augmentError(error, filename, files); }
};

const inlineCssUrl = async function ($, filename) {
	var files = [];
	try {
		var result;
		result = await inlineStyles($, filename);
		$ = result.$;
		files = R.concat(files, result.files);

		result = await inlineStyleAttributes($, filename);
		$ = result.$;
		files = R.concat(files, result.files);

		files = R.uniq(files);
		return { $, files };
	}
	catch (error) { throw augmentError(error, filename, files); }
};

module.exports = inlineCssUrl;
